"""Hive Device Attribute Module."""
from .helper.hive_data import Data
from .helper.hive_helper import HiveHelper as Helper
from .helper.logger import Logger


class Attributes:
    """Device Attributes Code."""

    def __init__(self):
        self.logger = Logger()
        self.type = "Attribute"

    async def state_attributes(self, n_id, _type):
        """Get HA State Attributes"""
        attr = {}

        if n_id in Data.products or n_id in Data.devices:
            attr.update({"available": (await self.online_offline(n_id))})
            if n_id in Data.BATTERY:
                battery = await self.battery(n_id)
                if battery is not None:
                    attr.update({"battery": str(battery) + "%"})
            if n_id in Data.MODE:
                attr.update({"mode": (await self.get_mode(n_id))})
        return attr

    async def online_offline(self, n_id):
        """Check if device is online"""
        state = None

        try:
            data = Data.devices[n_id]
            state = data["props"]["online"]
        except KeyError as e:
            await self.logger.error(e)

        return state

    async def get_mode(self, n_id):
        """Get sensor mode."""
        state = None
        final = None

        try:
            data = Data.products[n_id]
            state = data["state"]["mode"]
            final = Data.HIVETOHA[self.type].get(state, state)
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def battery(self, n_id):
        """Get device battery level."""
        state = None
        final = None

        try:
            data = Data.devices[n_id]
            state = data["props"]["battery"]
            final = state
            await self.logger.error_check(n_id, self.type, state)
        except KeyError as e:
            await self.logger.error(e)

        return final



"""Hive Action Module."""
from .helper.hive_data import Data
from .session import Session


class Action(Session):
    """Hive Action Code."""

    actionType = "Actions"

    async def get_action(self, device):
        """Get smart plug current power usage."""
        await self.logger.log(
            device["hiveID"], self.actionType, "Getting action data."
        )
        dev_data = {}

        if device["hiveID"] in Data.actions:
            dev_data = {
                "hiveID": device["hiveID"],
                "hiveName": device["hiveName"],
                "hiveType": device["hiveType"],
                "haName": device["haName"],
                "haType": device["haType"],
                "status": {"state": await self.get_state(device)},
                "power_usage": None,
                "deviceData": {},
                "custom": device.get("custom", None),
            }

            await self.logger.log(
                device["hiveID"],
                self.actionType,
                "action update {0}",
                info=[dev_data["status"]],
            )
            Data.ha_devices.update({device["hiveID"]: dev_data})
            return Data.ha_devices[device["hiveID"]]
        else:
            exists = Data.actions.get("hiveID", False)
            if exists is False:
                return "REMOVE"
            return device

    async def get_state(self, device):
        """Get action state."""
        final = None

        try:
            data = Data.actions[device["hiveID"]]
            final = data["enabled"]
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def turn_on(self, device):
        """Set action turn on."""
        import json

        final = False

        if device["hiveID"] in Data.actions:
            await self.hiveRefreshTokens()
            data = Data.actions[device["hiveID"]]
            data.update({"enabled": True})
            send = json.dumps(data)
            resp = await self.api.set_action(device["hiveID"], send)
            if resp["original"] == 200:
                final = True
                await self.getDevices(device["hiveID"])

        return final

    async def turn_off(self, device):
        """Set action to turn off."""
        import json

        final = False

        if device["hiveID"] in Data.actions:
            await self.hiveRefreshTokens()
            data = Data.actions[device["hiveID"]]
            data.update({"enabled": False})
            send = json.dumps(data)
            resp = await self.api.set_action(device["hiveID"], send)
            if resp["original"] == 200:
                final = True
                await self.getDevices(device["hiveID"])

        return final



import datetime
import operator

from .hive_data import Data


class HiveHelper:
    def __init__(self):
        """Hive Helper."""

    def getDeviceName(n_id):
        """Resolve a id into a name"""
        try:
            product_name = Data.products[n_id]["state"]["name"]
        except KeyError:
            product_name = False

        try:
            device_name = Data.devices[n_id]["state"]["name"]
        except KeyError:
            device_name = False

        if product_name:
            return product_name
        elif device_name:
            return device_name
        elif n_id == "No_ID":
            return "Hive"
        else:
            return n_id

    def deviceRecovered(self, n_id):
        """"Register that a device has recovered from being offline."""
        # name = HiveHelper.getDeviceName(n_id)
        if n_id in Data.errorList:
            Data.errorList.pop(n_id)

    def getDeviceFromID(self, n_id):
        """Get product/device data from ID"""
        data = False
        try:
            data = Data.ha_devices[n_id]
        except KeyError:
            pass

        return data

    def getDeviceData(self, product):
        """"Get device Data."""
        device = product
        type = product["type"]
        if type in ("heating", "hotwater"):
            for aDevice in Data.devices:
                if Data.devices[aDevice]["type"] in Data.HIVE_TYPES["Thermo"]:
                    try:
                        if (
                            product["props"]["zone"]
                            == Data.devices[aDevice]["props"]["zone"]
                        ):
                            device = Data.devices[aDevice]
                    except KeyError:
                        pass
        elif type == "trvcontrol":
            device = Data.devices[product["props"]["trvs"][0]]
        elif type == "warmwhitelight" and product["props"]["model"] == "SIREN001":
            device = Data.devices[product["parent"]]
        elif type == "sense":
            device = Data.devices[product["parent"]]
        else:
            device = Data.devices[product["id"]]

        return device

    def convertMinutesToTime(self, minutes_to_convert):
        """Convert minutes string to datetime."""
        hours_converted, minutes_converted = divmod(minutes_to_convert, 60)
        converted_time = datetime.strptime(
            str(hours_converted) + ":" + str(minutes_converted), "%H:%M"
        )
        converted_time_string = converted_time.strftime("%H:%M")
        return converted_time_string

    def getScheduleNNL(self, hive_api_schedule):
        """Get the schedule now, next and later of a given nodes schedule."""
        schedule_now_and_next = {}
        date_time_now = datetime.datetime.now()
        date_time_now_day_int = date_time_now.today().weekday()

        days_t = (
            "monday",
            "tuesday",
            "wednesday",
            "thursday",
            "friday",
            "saturday",
            "sunday",
        )

        days_rolling_list = list(days_t[date_time_now_day_int:] + days_t)[:7]

        full_schedule_list = []

        for day_index in range(0, len(days_rolling_list)):
            current_day_schedule = hive_api_schedule[days_rolling_list[day_index]]
            current_day_schedule_sorted = sorted(
                current_day_schedule,
                key=operator.itemgetter("start"),
                reverse=False,
            )

            for current_slot in range(0, len(current_day_schedule_sorted)):
                current_slot_custom = current_day_schedule_sorted[current_slot]

                slot_date = datetime.datetime.now() + datetime.timedelta(days=day_index)
                slot_time = self.convertMinutesToTime(current_slot_custom["start"])
                slot_time_date_s = slot_date.strftime("%d-%m-%Y") + " " + slot_time
                slot_time_date_dt = datetime.strptime(
                    slot_time_date_s, "%d-%m-%Y %H:%M"
                )
                if slot_time_date_dt <= date_time_now:
                    slot_time_date_dt = slot_time_date_dt + datetime.timedelta(days=7)

                current_slot_custom["Start_DateTime"] = slot_time_date_dt
                full_schedule_list.append(current_slot_custom)

        fsl_sorted = sorted(
            full_schedule_list,
            key=operator.itemgetter("Start_DateTime"),
            reverse=False,
        )

        schedule_now = fsl_sorted[-1]
        schedule_next = fsl_sorted[0]
        schedule_later = fsl_sorted[1]

        schedule_now["Start_DateTime"] = schedule_now[
            "Start_DateTime"
        ] - datetime.timedelta(days=7)

        schedule_now["End_DateTime"] = schedule_next["Start_DateTime"]
        schedule_next["End_DateTime"] = schedule_later["Start_DateTime"]
        schedule_later["End_DateTime"] = fsl_sorted[2]["Start_DateTime"]

        schedule_now_and_next["now"] = schedule_now
        schedule_now_and_next["next"] = schedule_next
        schedule_now_and_next["later"] = schedule_later

        return schedule_now_and_next



"""Hive Switch Module."""
from .helper.hive_data import Data
from .session import Session


class Plug(Session):
    """Hive Switch Code."""

    plugType = "Switch"

    async def get_plug(self, device):
        """Get smart plug data."""
        await self.logger.log(
            device["hiveID"], self.plugType, "Getting switch data."
        )
        device["deviceData"].update(
            {"online": await self.attr.online_offline(device["device_id"])}
        )
        dev_data = {}

        if device["deviceData"]["online"]:
            self.helper.deviceRecovered(device["device_id"])
            data = Data.devices[device["device_id"]]
            dev_data = {
                "hiveID": device["hiveID"],
                "hiveName": device["hiveName"],
                "hiveType": device["hiveType"],
                "haName": device["haName"],
                "haType": device["haType"],
                "device_id": device["device_id"],
                "device_name": device["device_name"],
                "status": {
                    "state": await self.get_state(device),
                    "power_usage": await self.get_power_usage(device),
                },
                "deviceData": data.get("props", None),
                "parentDevice": data.get("parent", None),
                "custom": device.get("custom", None),
                "attributes": await self.attr.state_attributes(
                    device["device_id"], device["hiveType"]
                ),
            }

            await self.logger.log(
                device["hiveID"],
                self.plugType,
                "Device update {0}",
                info=[dev_data["status"]],
            )
            Data.ha_devices.update({device["hiveID"]: dev_data})
            return Data.ha_devices[device["hiveID"]]
        else:
            await self.logger.error_check(
                device["device_id"], "ERROR", device["deviceData"]["online"]
            )
            return device

    async def get_state(self, device):
        """Get plug current state."""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["state"]["status"]
            final = Data.HIVETOHA["Switch"].get(state, state)
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def get_power_usage(self, device):
        """Get smart plug current power usage."""
        state = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["props"]["powerConsumption"]
        except KeyError as e:
            await self.logger.error(e)

        return state

    async def turn_on(self, device):
        """Set smart plug to turn on."""
        final = False

        if (
            device["hiveID"] in Data.products
            and device["deviceData"]["online"]
        ):
            await self.hiveRefreshTokens()
            data = Data.products[device["hiveID"]]
            resp = await self.api.set_state(
                data["type"], data["id"], status="ON"
            )
            if resp["original"] == 200:
                final = True
                await self.getDevices(device["hiveID"])

        return final

    async def turn_off(self, device):
        """Set smart plug to turn off."""
        final = False

        if (
            device["hiveID"] in Data.products
            and device["deviceData"]["online"]
        ):
            await self.hiveRefreshTokens()
            data = Data.products[device["hiveID"]]
            resp = await self.api.set_state(
                data["type"], data["id"], status="OFF"
            )
            if resp["original"] == 200:
                final = True
                await self.getDevices(device["hiveID"])

        return final



"""Hive Data Module."""
import datetime


class Data:
    """Hive Data"""

    # API Data
    products = {}
    devices = {}
    actions = {}
    user = {}
    ha_devices = {}
    MODE = []
    BATTERY = []
    HIVETOHA = {
        "Attribute": {True: "Online", False: "Offline"},
        "Boost": {None: "OFF", False: "OFF"},
        "Heating": {False: "OFF"},
        "Hotwater": {"MANUAL": "ON", None: "OFF", False: "OFF"},
        "Hub": {
            "Status": {True: 1, False: 0},
            "Smoke": {True: 1, False: 0},
            "Dog": {True: 1, False: 0},
            "Glass": {True: 1, False: 0},
        },
        "Light": {"ON": True, "OFF": False},
        "Sensor": {
            "OPEN": True,
            "CLOSED": False,
            True: "Online",
            False: "Offline",
        },
        "Switch": {"ON": True, "OFF": False},
    }
    HIVE_TYPES = {
        "Hub": ["hub", "sense"],
        "Thermo": ["thermostatui", "trv"],
        "Heating": ["heating", "trvcontrol"],
        "Hotwater": ["hotwater"],
        "Light": ["warmwhitelight", "tuneablelight", "colourtuneablelight"],
        "Sensor": ["motionsensor", "contactsensor"],
        "Switch": ["activeplug"],
    }
    sensor_commands = {
        "SMOKE_CO": "Hub.hub_smoke(Hub(), device)",
        "DOG_BARK": "Hub.hub_dog_bark(Hub(), device)",
        "GLASS_BREAK": "Hub.hub_glass(Hub(), device)",
        "CurrentTemperature": "Heating.current_temperature(Heating(), device)",
        "TargetTemperature": "Heating.target_temperature(Heating(), device)",
        "Heating_State": "Heating.get_state(Heating(), device)",
        "Heating_Mode": "Heating.get_mode(Heating(), device)",
        "Heating_Boost": "Heating.boost(Heating(), device)",
        "Hotwater_State": "Hotwater.get_state(Hotwater(), device)",
        "Hotwater_Mode": "Hotwater.get_mode(Hotwater(), device)",
        "Hotwater_Boost": "Hotwater.get_boost(Hotwater(), device)",
        "Battery": 'self.attr.battery(device["device_id"])',
        "Mode": 'self.attr.get_mode(device["hiveID"])',
        "Availability": "self.online(device)",
        "Connectivity": "self.online(device)",
    }

    # Session Data
    tokens = {}
    tokenCreated = datetime.datetime.now() - datetime.timedelta(seconds=4000)
    tokenExpiry = datetime.timedelta(seconds=1800)
    lastUpdate = datetime.datetime.now()
    intervalSeconds = datetime.timedelta(seconds=120)
    sensors = False
    file = False
    errorList = {}
    HttpCount = 0
    haWebsession = None
    loginData = {}

    # Platform data
    minMax = {}

    # Debugging data
    debugOutFolder = ""
    debugOutFile = ""
    debugEnabled = False
    debugList = []



"""Hive Sensor Module."""
from .heating import Heating  # noqa: F401
from .helper.hive_data import Data
from .hotwater import Hotwater  # noqa: F401
from .hub import Hub  # noqa: F401
from .session import Session


class Sensor(Session):
    """Hive Sensor Code."""

    sensorType = "Sensor"

    async def get_sensor(self, device):
        await self.logger.log(
            device["hiveID"], self.sensorType, "Getting sensor data."
        )
        device["deviceData"].update(
            {"online": await self.attr.online_offline(device["device_id"])}
        )
        data = {}

        if device["deviceData"]["online"] or device["hiveType"] in (
            "Availability",
            "Connectivity",
        ):
            if device["hiveType"] not in ("Availability", "Connectivity"):
                self.helper.deviceRecovered(device["device_id"])

            dev_data = {}
            dev_data = {
                "hiveID": device["hiveID"],
                "hiveName": device["hiveName"],
                "hiveType": device["hiveType"],
                "haName": device["haName"],
                "haType": device["haType"],
                "device_id": device.get("device_id", None),
                "device_name": device.get("device_name", None),
                "deviceData": {},
                "custom": device.get("custom", None),
            }

            if device["device_id"] in Data.devices:
                data = Data.devices.get(device["device_id"], {})
            elif device["hiveID"] in Data.products:
                data = Data.products.get(device["hiveID"], {})

            if (
                dev_data["hiveType"] in Data.sensor_commands
                or dev_data.get("custom", None) in Data.sensor_commands
            ):
                code = Data.sensor_commands.get(
                    dev_data["hiveType"],
                    Data.sensor_commands.get(dev_data["custom"]),
                )
                dev_data.update(
                    {
                        "status": {"state": await eval(code)},
                        "deviceData": data.get("props", None),
                        "parentDevice": data.get("parent", None),
                    }
                )
            elif device["hiveType"] in Data.HIVE_TYPES["Sensor"]:
                data = Data.devices.get(device["hiveID"], {})
                dev_data.update(
                    {
                        "status": {"state": await self.get_state(device)},
                        "deviceData": data.get("props", None),
                        "parentDevice": data.get("parent", None),
                        "attributes": await self.attr.state_attributes(
                            device["device_id"], device["hiveType"]
                        ),
                    }
                )

            await self.logger.log(
                device["hiveID"],
                self.sensorType,
                "Device update {0}",
                info=[dev_data["status"]],
            )
            Data.ha_devices.update({device["hiveID"]: dev_data})
            return Data.ha_devices[device["hiveID"]]
        else:
            await self.logger.error_check(
                device["device_id"], "ERROR", device["deviceData"]["online"]
            )
            return device

    async def get_state(self, device):
        """Get sensor state."""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            if data["type"] == "contactsensor":
                state = data["props"]["status"]
                final = Data.HIVETOHA[self.sensorType].get(state, state)
            elif data["type"] == "motionsensor":
                final = data["props"]["motion"]["status"]
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def online(self, device):
        """Get the online status of the Hive hub."""
        state = None
        final = None

        try:
            data = Data.devices[device["device_id"]]
            state = data["props"]["online"]
            final = Data.HIVETOHA[self.sensorType].get(state, state)
        except KeyError as e:
            await self.logger.error(e)

        return final



""""Hive Heating Module. """
from .helper.hive_data import Data
from .session import Session


class Heating(Session):
    """Hive Heating Code."""

    heatingType = "Heating"

    async def get_heating(self, device):
        await self.logger.log(
            device["hiveID"], self.heatingType, "Getting heating data."
        )
        device["deviceData"].update(
            {"online": await self.attr.online_offline(device["device_id"])}
        )

        if device["deviceData"]["online"]:
            dev_data = {}
            self.helper.deviceRecovered(device["device_id"])
            data = Data.devices[device["device_id"]]
            dev_data = {
                "hiveID": device["hiveID"],
                "hiveName": device["hiveName"],
                "hiveType": device["hiveType"],
                "haName": device["haName"],
                "haType": device["haType"],
                "device_id": device["device_id"],
                "device_name": device["device_name"],
                "temperatureunit": device["temperatureunit"],
                "min_temp": await self.min_temperature(device),
                "max_temp": await self.max_temperature(device),
                "status": {
                    "current_temperature": await self.current_temperature(
                        device
                    ),
                    "target_temperature": await self.target_temperature(
                        device
                    ),
                    "action": await self.current_operation(device),
                    "mode": await self.get_mode(device),
                    "boost": await self.boost(device),
                },
                "deviceData": data.get("props", None),
                "parentDevice": data.get("parent", None),
                "custom": device.get("custom", None),
                "attributes": await self.attr.state_attributes(
                    device["device_id"], device["hiveType"]
                ),
            }
            await self.logger.log(
                device["hiveID"],
                self.heatingType,
                "Device update {0}",
                info=[dev_data["status"]],
            )
            Data.ha_devices.update({device["hiveID"]: dev_data})
            return Data.ha_devices[device["hiveID"]]
        else:
            await self.logger.error_check(
                device["device_id"], "ERROR", device["deviceData"]["online"]
            )
            return device

    @staticmethod
    async def min_temperature(device):
        """Get heating minimum target temperature."""
        if device["hiveType"] == "nathermostat":
            return Data.products[device["hiveID"]]["props"]["minHeat"]
        return 5

    @staticmethod
    async def max_temperature(device):
        """Get heating maximum target temperature."""
        if device["hiveType"] == "nathermostat":
            return Data.products[device["hiveID"]]["props"]["maxHeat"]
        return 32

    async def current_temperature(self, device):
        """Get heating current temperature."""
        from datetime import datetime

        f_state = None
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["props"]["temperature"]

            if device["hiveID"] in Data.minMax:
                if Data.minMax[device["hiveID"]]["TodayDate"] != datetime.date(
                    datetime.now()
                ):
                    Data.minMax[device["hiveID"]]["TodayMin"] = 1000
                    Data.minMax[device["hiveID"]]["TodayMax"] = -1000
                    Data.minMax[device["hiveID"]]["TodayDate"] = datetime.date(
                        datetime.now()
                    )

                    if state < Data.minMax[device["hiveID"]]["TodayMin"]:
                        Data.minMax[device["hiveID"]]["TodayMin"] = state

                    if state > Data.minMax[device["hiveID"]]["TodayMax"]:
                        Data.minMax[device["hiveID"]]["TodayMax"] = state

                    if state < Data.minMax[device["hiveID"]]["RestartMin"]:
                        Data.minMax[device["hiveID"]]["RestartMin"] = state

                    if state > Data.minMax[device["hiveID"]]["RestartMax"]:
                        Data.minMax[device["hiveID"]]["RestartMax"] = state
            else:
                data = {
                    "TodayMin": state,
                    "TodayMax": state,
                    "TodayDate": str(datetime.date(datetime.now())),
                    "RestartMin": state,
                    "RestartMax": state,
                }
                Data.minMax[device["hiveID"]] = data
            f_state = round(float(state), 1)
            await self.logger.log(
                device["hiveID"],
                self.heatingType + "_Extra",
                "Current Temp is {0}",
                info=[str(state)],
            )

            final = f_state
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def minmax_temperatures(self, device):
        """Min/Max Temp"""
        state = None
        final = None

        try:
            state = Data.minMax[device["hiveID"]]
            final = state
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def target_temperature(self, device):
        """Get heating target temperature."""
        state = None

        try:
            data = Data.products[device["hiveID"]]
            state = float(data["state"].get("target", None))
            state = float(data["state"].get("heat", state))
        except (KeyError, TypeError) as e:
            await self.logger.error(e)

        return state

    async def get_mode(self, device):
        """Get heating current mode."""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["state"]["mode"]
            if state == "BOOST":
                state = data["props"]["previous"]["mode"]
            final = Data.HIVETOHA[self.heatingType].get(state, state)
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def get_state(self, device):
        """Get heating current state."""
        state = None
        final = None

        try:
            current_temp = await self.current_temperature(device)
            target_temp = await self.target_temperature(device)
            if current_temp < target_temp:
                state = "ON"
            else:
                state = "OFF"
            final = Data.HIVETOHA[self.heatingType].get(state, state)
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def current_operation(self, device):
        """Get heating current operation."""
        state = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["props"]["working"]
        except KeyError as e:
            await self.logger.error(e)

        return state

    async def boost(self, device):
        """Get heating boost current status."""
        state = None

        try:
            data = Data.products[device["hiveID"]]
            state = Data.HIVETOHA["Boost"].get(
                data["state"].get("boost", False), "ON"
            )
        except KeyError as e:
            await self.logger.error(e)

        return state

    async def get_boost_time(self, device):
        """Get heating boost time remaining."""
        if await self.boost(device) == "ON":
            state = None

            try:
                data = Data.products[device["hiveID"]]
                state = data["state"]["boost"]
            except KeyError as e:
                await self.logger.error(e)

            return state
        return None

    @staticmethod
    async def get_operation_modes():
        """Get heating list of possible modes."""
        return ["SCHEDULE", "MANUAL", "OFF"]

    async def get_schedule_now_next_later(self, device):
        """Hive get heating schedule now, next and later."""
        online = await self.attr.online_offline(device["device_id"])
        current_mode = await self.get_mode(device)
        state = None

        try:
            if online and current_mode == "SCHEDULE":
                data = Data.products[device["hiveID"]]
                state = await self.helper.getScheduleNNL(
                    data["state"]["schedule"]
                )
        except KeyError as e:
            await self.logger.error(e)

        return state

    async def set_target_temperature(self, device, new_temp):
        """Set heating target temperature."""
        await self.hiveRefreshTokens()
        final = False

        if (
            device["hiveID"] in Data.products
            and device["deviceData"]["online"]
        ):
            await self.hiveRefreshTokens()
            data = Data.products[device["hiveID"]]
            resp = await self.api.set_state(
                data["type"], device["hiveID"], target=new_temp
            )

            if resp["original"] == 200:
                await self.getDevices(device["hiveID"])
                final = True

        return final

    async def set_mode(self, device, new_mode):
        """Set heating mode."""
        await self.hiveRefreshTokens()
        final = False

        if (
            device["hiveID"] in Data.products
            and device["deviceData"]["online"]
        ):
            data = Data.products[device["hiveID"]]
            resp = await self.api.set_state(
                data["type"], device["hiveID"], mode=new_mode
            )

            if resp["original"] == 200:
                await self.getDevices(device["hiveID"])
                final = True

        return final

    async def turn_boost_on(self, device, mins, temp):
        """Turn heating boost on."""

        if mins > 0 and temp >= await self.min_temperature(device):
            if temp <= await self.max_temperature(device):
                await self.hiveRefreshTokens()
                final = False

                if (
                    device["hiveID"] in Data.products
                    and device["deviceData"]["online"]
                ):
                    data = Data.products[device["hiveID"]]
                    resp = await self.api.set_state(
                        data["type"],
                        device["hiveID"],
                        mode="BOOST",
                        boost=mins,
                        target=temp,
                    )

                    if resp["original"] == 200:
                        await self.getDevices(device["hiveID"])
                        final = True

                return final
        return None

    async def turn_boost_off(self, device):
        """Turn heating boost off."""
        final = False

        if (
            device["hiveID"] in Data.products
            and device["deviceData"]["online"]
        ):
            await self.hiveRefreshTokens()
            data = Data.products[device["hiveID"]]
            await self.getDevices(device["hiveID"])
            if await self.boost(device) == "ON":
                prev_mode = data["props"]["previous"]["mode"]
                if prev_mode == "MANUAL" or prev_mode == "OFF":
                    pre_temp = data["props"]["previous"].get("target", 7)
                    resp = await self.api.set_state(
                        data["type"],
                        device["hiveID"],
                        mode=prev_mode,
                        target=pre_temp,
                    )
                else:
                    resp = await self.api.set_state(
                        data["type"], device["hiveID"], mode=prev_mode
                    )
                if resp["original"] == 200:
                    await self.getDevices(device["hiveID"])
                    final = True

        return final



import logging
import sys


class DebugContext:
    """ Debug context to trace any function calls inside the context """

    def __init__(self, name, enabled):
        self.name = name
        self.enabled = enabled
        self.logging = logging.getLogger(__name__)

    def __enter__(self):
        print("Entering Debug Decorated func")
        # Set the trace function to the trace_calls function
        # So all events are now traced
        self.trace_calls

    def trace_calls(self, frame, event, arg):
        # We want to only trace our call to the decorated function
        if event != "call":
            return
        elif frame.f_code.co_name != self.name:
            return
        # return the trace function to use when you go into that
        # function call
        return self.trace_lines

    def trace_lines(self, frame, event, arg):
        # If you want to print local variables each line
        # keep the check for the event 'line'
        # If you want to print local variables only on return
        # check only for the 'return' event
        if event not in ["line", "return"]:
            return
        co = frame.f_code
        func_name = co.co_name
        line_no = frame.f_lineno
        filename = co.co_filename
        local_vars = frame.f_locals
        text = f"  {func_name} {event} {line_no} locals: {local_vars}"
        self.logging.debug(text)


def debug_decorator(enabled=False):
    """ Debug decorator to call the function within the debug context """

    def decorated_func(func):
        def wrapper(*args, **kwargs):
            with DebugContext(func.__name__, enabled):
                return_value = func(*args, **kwargs)
            return return_value

        return wrapper

    return decorated_func



"""__init__.py"""
from .api.hive_api import HiveApi
from .api.hive_async_api import HiveAsync
from .api.hive_auth import HiveAuth
from .api.hive_auth_async import HiveAuthAsync
from .helper.const import *
from .helper.hive_exceptions import *
from .helper.hive_helper import HiveHelper
from .hive import Hive
from .session import Session



"""Constants for Pyhiveapi."""
PACKAGE_NAME = "Pyhiveapi"
PACKAGE_DIR = "/pyhiveapi/"
SMS_REQUIRED = "SMS_MFA"


# HTTP return codes.
HTTP_OK = 200
HTTP_CREATED = 201
HTTP_ACCEPTED = 202
HTTP_MOVED_PERMANENTLY = 301
HTTP_BAD_REQUEST = 400
HTTP_UNAUTHORIZED = 401
HTTP_FORBIDDEN = 403
HTTP_NOT_FOUND = 404
HTTP_METHOD_NOT_ALLOWED = 405
HTTP_UNPROCESSABLE_ENTITY = 422
HTTP_TOO_MANY_REQUESTS = 429
HTTP_INTERNAL_SERVER_ERROR = 500
HTTP_BAD_GATEWAY = 502
HTTP_SERVICE_UNAVAILABLE = 503



"""Hive Light Module."""
import colorsys

from .helper.hive_data import Data
from .session import Session


class Light(Session):
    """Hive Light Code."""

    lightType = "Light"

    async def get_light(self, device):
        """Get light data."""
        await self.logger.log(
            device["hiveID"], self.lightType, "Getting light data."
        )
        device["deviceData"].update(
            {"online": await self.attr.online_offline(device["device_id"])}
        )
        dev_data = {}

        if device["deviceData"]["online"]:
            self.helper.deviceRecovered(device["device_id"])
            data = Data.devices[device["device_id"]]
            dev_data = {
                "hiveID": device["hiveID"],
                "hiveName": device["hiveName"],
                "hiveType": device["hiveType"],
                "haName": device["haName"],
                "haType": device["haType"],
                "device_id": device["device_id"],
                "device_name": device["device_name"],
                "status": {
                    "state": await self.get_state(device),
                    "brightness": await self.get_brightness(device),
                },
                "deviceData": data.get("props", None),
                "parentDevice": data.get("parent", None),
                "custom": device.get("custom", None),
                "attributes": await self.attr.state_attributes(
                    device["device_id"], device["hiveType"]
                ),
            }

            if device["hiveType"] in ("tuneablelight", "colourtuneablelight"):
                dev_data.update(
                    {
                        "min_mireds": await self.get_min_color_temp(device),
                        "max_mireds": await self.get_max_color_temp(device),
                    }
                )
                dev_data["status"].update(
                    {"color_temp": await self.get_color_temp(device)}
                )
            if device["hiveType"] == "colourtuneablelight":
                dev_data["status"].update(
                    {
                        "hs_color": await self.get_color(device),
                        "mode": await self.getColourMode(device),
                    }
                )

            await self.logger.log(
                device["hiveID"],
                self.lightType,
                "Device update {0}",
                info=[dev_data["status"]],
            )
            Data.ha_devices.update({device["hiveID"]: dev_data})
            return Data.ha_devices[device["hiveID"]]
        else:
            await self.logger.error_check(
                device["device_id"], "ERROR", device["deviceData"]["online"]
            )
            return device

    async def get_state(self, device):
        """Get light current state."""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["state"]["status"]
            final = Data.HIVETOHA[self.lightType].get(state, state)
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def get_brightness(self, device):
        """Get light current brightness."""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["state"]["brightness"]
            final = (state / 100) * 255
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def get_min_color_temp(self, device):
        """Get light minimum color temperature."""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["props"]["colourTemperature"]["max"]
            final = round((1 / state) * 1000000)
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def get_max_color_temp(self, device):
        """Get light maximum color temperature."""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["props"]["colourTemperature"]["min"]
            final = round((1 / state) * 1000000)
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def get_color_temp(self, device):
        """Get light current color temperature."""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["state"]["colourTemperature"]
            final = round((1 / state) * 1000000)
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def get_color(self, device):
        """Get light current colour"""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = [
                (data["state"]["hue"]) / 360,
                (data["state"]["saturation"]) / 100,
                (data["state"]["value"]) / 100,
            ]
            final = tuple(
                int(i * 255)
                for i in colorsys.hsv_to_rgb(state[0], state[1], state[2])
            )
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def getColourMode(self, device):
        """Get Colour Mode."""
        state = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["state"]["colourMode"]
        except KeyError as e:
            await self.logger.error(e)

        return state

    async def turn_off(self, device):
        """Set light to turn off."""
        final = False

        if (
            device["hiveID"] in Data.products
            and device["deviceData"]["online"]
        ):
            await self.hiveRefreshTokens()
            data = Data.products[device["hiveID"]]
            resp = await self.api.set_state(
                data["type"], device["hiveID"], status="OFF"
            )

            if resp["original"] == 200:
                final = True
                await self.getDevices(device["hiveID"])

        return final

    async def turn_on(self, device, brightness, color_temp, color):
        """Set light to turn on."""
        final = False

        if (
            device["hiveID"] in Data.products
            and device["deviceData"]["online"]
        ):
            await self.hiveRefreshTokens()
            data = Data.products[device["hiveID"]]

            if brightness is not None:
                return await self.set_brightness(device, brightness)
            if color_temp is not None:
                return await self.set_color_temp(device, color_temp)
            if color is not None:
                return await self.set_color(device, color)

            resp = await self.api.set_state(
                data["type"], device["hiveID"], status="ON"
            )
            if resp["original"] == 200:
                final = True
                await self.getDevices(device["hiveID"])

        return final

    async def set_brightness(self, device, n_brightness):
        """Set brightness of the light."""
        final = False

        if (
            device["hiveID"] in Data.products
            and device["deviceData"]["online"]
        ):
            await self.hiveRefreshTokens()
            data = Data.products[device["hiveID"]]
            resp = await self.api.set_state(
                data["type"],
                device["hiveID"],
                status="ON",
                brightness=n_brightness,
            )
            if resp["original"] == 200:
                final = True
                await self.getDevices(device["hiveID"])

        return final

    async def set_color_temp(self, device, color_temp):
        """Set light to turn on."""
        final = False

        if (
            device["hiveID"] in Data.products
            and device["deviceData"]["online"]
        ):
            await self.hiveRefreshTokens()
            data = Data.products[device["hiveID"]]

            if data["type"] == "tuneablelight":
                resp = await self.api.set_state(
                    data["type"],
                    device["hiveID"],
                    colourTemperature=color_temp,
                )
            else:
                resp = await self.api.set_state(
                    data["type"],
                    device["hiveID"],
                    colourMode="WHITE",
                    colourTemperature=color_temp,
                )

            if resp["original"] == 200:
                final = True
                await self.getDevices(device["hiveID"])

        return final

    async def set_color(self, device, new_color):
        """Set light to turn on."""
        final = False

        if (
            device["hiveID"] in Data.products
            and device["deviceData"]["online"]
        ):
            await self.hiveRefreshTokens()
            data = Data.products[device["hiveID"]]

            resp = await self.api.set_state(
                data["type"],
                device["hiveID"],
                colourMode="COLOUR",
                hue=str(new_color[0]),
                saturation=str(new_color[1]),
                value=str(new_color[2]),
            )
            if resp["original"] == 200:
                final = True
                await self.getDevices(device["hiveID"])

        return final



""" Hive Session Module."""
import asyncio
import copy
import json
import operator
import os
import time
from datetime import datetime, timedelta

from aiohttp.web import HTTPException

from .api.hive_async_api import HiveAsync
from .device_attributes import Attributes
from .helper.debugger import debug_decorator
from .helper.hive_data import Data
from .helper.hive_exceptions import HiveApiError
from .helper.hive_helper import HiveHelper
from .helper.logger import Logger


class Session:
    """Hive Session Code"""

    sessionType = "Session"

    def __init__(self):
        """Initialise the base variable values."""
        self.updateLock = asyncio.Lock()
        self.api = HiveAsync()
        self.logger = Logger()
        self.helper = HiveHelper()
        self.attr = Attributes()
        self.devices = {}

    async def openFile(self, file):
        path = os.path.dirname(os.path.realpath(__file__)) + "/helper/" + file
        with open(path, "r") as j:
            data = json.loads(j.read())

        return data

    async def add_list(self, type, data, **kwargs):
        """Add entity to the list"""
        add = False if kwargs.get("custom") and not Data.sensors else True
        device = self.helper.getDeviceData(data)
        device_name = (
            device["state"]["name"]
            if device["state"]["name"] != "Receiver"
            else "Heating"
        )
        formatted_data = {}

        if add:
            try:
                formatted_data = {
                    "hiveID": data.get("id", ""),
                    "hiveName": device_name,
                    "hiveType": data.get("type", ""),
                    "haType": type,
                    "deviceData": device.get("props", data.get("props", {})),
                    "parentDevice": data.get("parent", None),
                    "isGroup": data.get("isGroup", False),
                    "device_id": device["id"],
                    "device_name": device_name,
                }

                if kwargs.get("haName", "FALSE")[0] == " ":
                    kwargs["haName"] = device_name + kwargs["haName"]
                else:
                    formatted_data["haName"] = device_name
                formatted_data.update(kwargs)
            except KeyError as e:
                await self.logger.error(e)

            self.devices[type].append(formatted_data)
        return add

    async def updateInterval(self, new_interval):
        "Update the scan interval."
        interval = timedelta(seconds=new_interval)
        if interval < timedelta(seconds=15):
            interval = timedelta(seconds=15)
        Data.intervalSeconds = interval
        text = "Scan interval updated to " + str(Data.intervalSeconds)
        await self.logger.log("No_ID", self.sessionType, text)

    async def useFile(self, username=None):
        "Update to check if file is being used."
        using_file = True if username == "use@file.com" else False
        if using_file:
            Data.file = True
            text = "Using a file."
            await self.logger.log("No_ID", self.sessionType, text)

    async def hiveRefreshTokens(self):
        """Refresh Hive tokens."""
        updated = False

        if Data.file:
            await self.logger.log(
                "No_ID",
                self.sessionType,
                "useFile is active - Cannot refresh tokens.",
            )
            return None
        else:
            await self.logger.log(
                "No_ID",
                self.sessionType,
                "Checking if new tokens are required",
            )

            expiry_time = Data.tokenCreated + Data.tokenExpiry
            if datetime.now() >= expiry_time:
                await self.logger.log(
                    "No_ID", self.sessionType, "Attempting to refresh tokens."
                )
                updated = await self.api.refreshTokens()

        return updated

    async def updateData(self, device):
        """Get latest data for Hive nodes - rate limiting."""
        await self.logger.checkDebugging(Data.debugList)
        await self.updateLock.acquire()
        updated = False
        try:
            ep = Data.lastUpdate + Data.intervalSeconds
            if datetime.now() >= ep:
                await self.getDevices(device["hiveID"])
                updated = True
        finally:
            self.updateLock.release()

        return updated

    async def getDevices(self, n_id):
        """Get latest data for Hive nodes."""
        get_nodes_successful = False
        api_resp_d = None

        try:
            await asyncio.sleep(1)
            if Data.file:
                await self.logger.log(n_id, "API", "Using file")
                api_resp_d = await self.openFile("data.json")
            elif Data.tokens is not None:
                await self.logger.log(n_id, "Session", "Getting hive device info")
                await self.hiveRefreshTokens()
                api_resp_d = await self.api.getAll()
                if operator.contains(str(api_resp_d["original"]), "20") is False:
                    raise HTTPException
                elif api_resp_d["parsed"] is None:
                    raise HiveApiError

            api_resp_p = api_resp_d["parsed"]
            tmpProducts = {}
            tmpDevices = {}
            tmpActions = {}

            for hiveType in api_resp_p:
                if hiveType == "user":
                    Data.user = api_resp_p[hiveType]
                if hiveType == "products":
                    for aProduct in api_resp_p[hiveType]:
                        tmpProducts.update({aProduct["id"]: aProduct})
                if hiveType == "devices":
                    for aDevice in api_resp_p[hiveType]:
                        tmpDevices.update({aDevice["id"]: aDevice})
                if hiveType == "actions":
                    for aAction in api_resp_p[hiveType]:
                        tmpActions.update({aAction["id"]: aAction})

            if len(tmpProducts) > 0:
                Data.products = copy.deepcopy(tmpProducts)
            if len(tmpDevices) > 0:
                Data.devices = copy.deepcopy(tmpDevices)
            Data.actions = copy.deepcopy(tmpActions)
            Data.lastUpdate = datetime.now()
            get_nodes_successful = True
        except (
            IOError,
            RuntimeError,
            HiveApiError,
            ConnectionError,
            HTTPException,
        ):
            await self.logger.log("No_ID", "API", "Api didn't receive any data")
            get_nodes_successful = False

        return get_nodes_successful

    async def startSession(self, config):
        """Setup the Hive platform."""
        Data.sensors = config.get("add_sensors", False)
        await self.logger.checkDebugging(config["options"].get("debug", []))
        await self.logger.log("No_ID", self.sessionType, "Initialising Hive Component.")
        await self.updateInterval(config["options"]["scan_interval"])
        await self.useFile(config.get("username", False))

        if config["tokens"] is not None and not Data.file:
            await self.logger.log("No_ID", self.sessionType, "Logging into Hive.")
            Data.tokens.update({"token": config["tokens"]["IdToken"]})
            Data.tokens.update({"refreshToken": config["tokens"]["RefreshToken"]})
            Data.tokens.update({"accessToken": config["tokens"]["AccessToken"]})
        elif Data.file:
            await self.logger.log(
                "No_ID",
                self.sessionType,
                "Loading up a hive session with a preloaded file.",
            )
        else:
            return "UNKNOWN_CONFIGURATION"

        try:
            await self.getDevices("No_ID")
        except HTTPException:
            return HTTPException

        if Data.devices == {} or Data.products == {}:
            await self.logger.log("No_ID", self.sessionType, "Failed to get data")
            return "INVALID_REAUTH"

        await self.logger.log("No_ID", self.sessionType, "Creating list of devices")
        self.devices["binary_sensor"] = []
        self.devices["climate"] = []
        self.devices["light"] = []
        self.devices["sensor"] = []
        self.devices["switch"] = []
        self.devices["water_heater"] = []

        for aProduct in Data.products:
            p = Data.products[aProduct]
            if p.get("isGroup", False):
                continue

            if Data.products[aProduct]["type"] == "sense":
                await self.add_list(
                    "binary_sensor",
                    p,
                    haName="Glass Detection",
                    hiveType="GLASS_BREAK",
                )
                await self.add_list(
                    "binary_sensor",
                    p,
                    haName="Smoke Detection",
                    hiveType="SMOKE_CO",
                )
                await self.add_list(
                    "binary_sensor",
                    p,
                    haName="Dog Bark Detection",
                    hiveType="DOG_BARK",
                )

            if Data.products[aProduct]["type"] in Data.HIVE_TYPES["Heating"]:

                Data.MODE.append(p["id"])
                await self.add_list(
                    "climate",
                    p,
                    temperatureunit=Data.user["temperatureUnit"],
                )
                await self.add_list(
                    "sensor",
                    p,
                    haName=" Current Temperature",
                    hiveType="CurrentTemperature",
                    custom=True,
                )
                await self.add_list(
                    "sensor",
                    p,
                    haName=" Target Temperature",
                    hiveType="TargetTemperature",
                    custom=True,
                )
                await self.add_list(
                    "sensor",
                    p,
                    haName=" State",
                    hiveType="Heating_State",
                    custom=True,
                )
                await self.add_list(
                    "sensor",
                    p,
                    haName=" Mode",
                    hiveType="Heating_Mode",
                    custom=True,
                )
                await self.add_list(
                    "sensor",
                    p,
                    haName=" Boost",
                    hiveType="Heating_Boost",
                    custom=True,
                )

            if Data.products[aProduct]["type"] in Data.HIVE_TYPES["Hotwater"]:
                await self.add_list("water_heater", p)
                await self.add_list(
                    "sensor",
                    p,
                    haName="Hotwater State",
                    hiveType="Hotwater_State",
                    custom=True,
                )
                await self.add_list(
                    "sensor",
                    p,
                    haName="Hotwater Mode",
                    hiveType="Hotwater_Mode",
                    custom=True,
                )
                await self.add_list(
                    "sensor",
                    p,
                    haName="Hotwater Boost",
                    hiveType="Hotwater_Boost",
                    custom=True,
                )

            if Data.products[aProduct]["type"] in Data.HIVE_TYPES["Switch"]:
                Data.MODE.append(p["id"])
                await self.add_list("switch", p)
                await self.add_list(
                    "sensor",
                    p,
                    haName=" Mode",
                    hiveType="Mode",
                    custom=True,
                )
                await self.add_list(
                    "sensor",
                    p,
                    haName=" Availability",
                    hiveType="Availability",
                    custom=True,
                )

            if Data.products[aProduct]["type"] in Data.HIVE_TYPES["Light"]:
                Data.MODE.append(p["id"])
                await self.add_list("light", p)
                await self.add_list(
                    "sensor",
                    p,
                    haName=" Mode",
                    hiveType="Mode",
                    custom=True,
                )
                await self.add_list(
                    "sensor",
                    p,
                    haName=" Availability",
                    hiveType="Availability",
                    custom=True,
                )

            if Data.products[aProduct]["type"] in Data.HIVE_TYPES["Sensor"]:
                await self.add_list("binary_sensor", p)

        for aDevice in Data.devices:
            d = Data.devices[aDevice]
            if (
                Data.devices[aDevice]["type"] in Data.HIVE_TYPES["Thermo"]
                or Data.devices[aDevice]["type"] in Data.HIVE_TYPES["Sensor"]
            ):
                Data.BATTERY.append(d["id"])
                await self.add_list(
                    "sensor",
                    d,
                    haName=" Battery Level",
                    hiveType="Battery",
                )
                await self.add_list(
                    "sensor",
                    d,
                    haName=" Availability",
                    hiveType="Availability",
                    custom=True,
                )

            if Data.devices[aDevice]["type"] in Data.HIVE_TYPES["Hub"]:
                await self.add_list(
                    "binary_sensor",
                    d,
                    haName="Hive Hub Status",
                    hiveType="Connectivity",
                )

        if "action" in Data.HIVE_TYPES["Switch"]:
            for action in Data.actions:
                a = Data.actions[action]
                await self.add_list(
                    "switch",
                    a,
                    hiveName=a["name"],
                    haName=a["name"],
                    hiveType="action",
                )

        await self.logger.log(
            "No_ID", self.sessionType, "Hive component has initialised"
        )

        return self.devices

    @staticmethod
    async def epochtime(date_time, pattern, action):
        """ date/time conversion to epoch"""
        if action == "to_epoch":
            pattern = "%d.%m.%Y %H:%M:%S"
            epochtime = int(time.mktime(time.strptime(str(date_time), pattern)))
            return epochtime
        elif action == "from_epoch":
            date = datetime.fromtimestamp(int(date_time)).strftime(pattern)
            return date



"""Custom Logging Module."""
import inspect
import logging
from datetime import datetime

from .hive_data import Data
from .hive_helper import HiveHelper


class Logger:
    """Custom Logging Code."""

    LOGGER = logging.getLogger(__name__)

    @staticmethod
    async def checkDebugging(enable_debug: list):
        """Check Logging Active"""

        if len(enable_debug) > 0:
            Data.debugEnabled = True
        else:
            Data.debugEnabled = False

        Data.debugList = enable_debug
        return Data.debugEnabled

    async def log(self, n_id, l_type, new_message, **kwargs):
        """Output new log entry if logging is turned on."""
        name = HiveHelper.getDeviceName(n_id) + " - "
        data = kwargs.get("info", [])
        if "_" in l_type:
            nxt = l_type.split("_")
            for i in nxt:
                if i != "Extra":
                    if i in Data.debugList and "Extra" in Data.debugList:
                        l_type = i
                        break

        if (
            Data.debugEnabled
            and new_message is not None
            and any(elem in Data.debugList for elem in [l_type, "All"])
        ):
            if l_type != "ERROR":
                logging_data = name + new_message.format(*data)
                self.LOGGER.debug(logging_data)

            try:
                l_file = open(Data.debugOutFile, "a")
                l_file.write(
                    datetime.now().strftime("%d-%b-%Y %H:%M:%S")
                    + " - "
                    + l_type
                    + " - "
                    + name
                    + " : "
                    + new_message.format(*data)
                    + "\n"
                )
                l_file.close()
            except FileNotFoundError:
                pass
        else:
            pass

    async def error(self, e="UNKNOWN"):
        """An unexpected error has occured"""
        self.LOGGER.error(
            f"An unexpected error has occured whilst executing {inspect.stack()[1][3]}"
            f" with exception {e.__class__} {e}"
        )

    async def error_check(self, n_id, n_type, error_type, **kwargs):
        """Error has occurred."""
        message = None
        new_data = None
        result = False
        name = HiveHelper.getDeviceName(n_id)

        if error_type is False:
            message = "Device offline could not update entity - " + name
            result = True
            if n_id not in Data.errorList:
                self.LOGGER.warning(message)
                Data.errorList.update({n_id: datetime.now()})
        elif error_type == "Failed":
            message = "ERROR - No data found for device - " + name
            result = True
            if n_id not in Data.errorList:
                self.LOGGER.error(message)
                Data.errorList.update({n_id: datetime.now()})

        await self.log(n_id, n_type, message, info=[new_data])
        return result



import asyncio
import base64
import binascii
import concurrent.futures
import datetime
import functools
import hashlib
import hmac
import os
import re

import boto3
import botocore
import six

from .hive_api import HiveApi

# https://github.com/aws/amazon-cognito-identity-js/blob/master/src/AuthenticationHelper.js#L22
n_hex = (
    "FFFFFFFFFFFFFFFFC90FDAA22168C234C4C6628B80DC1CD1"
    + "29024E088A67CC74020BBEA63B139B22514A08798E3404DD"
    + "EF9519B3CD3A431B302B0A6DF25F14374FE1356D6D51C245"
    + "E485B576625E7EC6F44C42E9A637ED6B0BFF5CB6F406B7ED"
    + "EE386BFB5A899FA5AE9F24117C4B1FE649286651ECE45B3D"
    + "C2007CB8A163BF0598DA48361C55D39A69163FA8FD24CF5F"
    + "83655D23DCA3AD961C62F356208552BB9ED529077096966D"
    + "670C354E4ABC9804F1746C08CA18217C32905E462E36CE3B"
    + "E39E772C180E86039B2783A2EC07A28FB5C55DF06F4C52C9"
    + "DE2BCBF6955817183995497CEA956AE515D2261898FA0510"
    + "15728E5A8AAAC42DAD33170D04507A33A85521ABDF1CBA64"
    + "ECFB850458DBEF0A8AEA71575D060C7DB3970F85A6E1E4C7"
    + "ABF5AE8CDB0933D71E8C94E04A25619DCEE3D2261AD2EE6B"
    + "F12FFA06D98A0864D87602733EC86A64521F2B18177B200C"
    + "BBE117577A615D6C770988C0BAD946E208E24FA074E5AB31"
    + "43DB5BFCE0FD108E4B82D120A93AD2CAFFFFFFFFFFFFFFFF"
)
# https://github.com/aws/amazon-cognito-identity-js/blob/master/src/AuthenticationHelper.js#L49
g_hex = "2"
info_bits = bytearray("Caldera Derived Key", "utf-8")
pool = concurrent.futures.ThreadPoolExecutor()


class HiveAuthAsync(object):

    NEW_PASSWORD_REQUIRED_CHALLENGE = "NEW_PASSWORD_REQUIRED"
    PASSWORD_VERIFIER_CHALLENGE = "PASSWORD_VERIFIER"
    SMS_MFA_CHALLENGE = "SMS_MFA"

    def __init__(
        self, username, password, pool_region=None, client_secret=None
    ):
        if pool_region is not None:
            raise ValueError(
                "pool_region and client should not both be specified "
                "(region should be passed to the boto3 client instead)"
            )

        self.loop = asyncio.get_event_loop()
        self.username = username
        self.password = password
        self.api = HiveApi()
        self.user_id = "user_id"
        self.client_secret = client_secret
        self.big_n = hex_to_long(n_hex)
        self.g = hex_to_long(g_hex)
        self.k = hex_to_long(hex_hash("00" + n_hex + "0" + g_hex))
        self.small_a_value = self.generate_random_small_a()
        self.large_a_value = self.calculate_a()
        self.useFile = True if self.username == "use@file.com" else False
        self.file_response = {"AuthenticationResult": {"AccessToken": "file"}}

    async def async_init(self):
        self.data = self.loop.run_in_executor(None, self.api.getLoginInfo)
        await self.data
        self.__pool_id = self.data._result.get("UPID")
        self.__client_id = self.data._result.get("CLIID")
        self.__region = self.data._result.get("REGION").split("_")[0]
        self.client = self.loop.run_in_executor(
            None, boto3.client, "cognito-idp", self.__region
        )
        await self.client

    def generate_random_small_a(self):
        """
        helper function to generate a random big integer
        :return {Long integer} a random value.
        """
        random_long_int = get_random(128)
        return random_long_int % self.big_n

    def calculate_a(self):
        """
        Calculate the client's public value A = g^a%N
        with the generated random number a
        :param {Long integer} a Randomly generated small A.
        :return {Long integer} Computed large A.
        """
        big_a = pow(self.g, self.small_a_value, self.big_n)
        # safety check
        if (big_a % self.big_n) == 0:
            raise ValueError("Safety check for A failed")
        return big_a

    def get_password_authentication_key(
        self, username, password, server_b_value, salt
    ):
        """
        Calculates the final hkdf based on computed S value, \
            and computed U value and the key
        :param {String} username Username.
        :param {String} password Password.
        :param {Long integer} server_b_value Server B value.
        :param {Long integer} salt Generated salt.
        :return {Buffer} Computed HKDF value.
        """
        server_b_value = hex_to_long(server_b_value)
        u_value = calculate_u(self.large_a_value, server_b_value)
        if u_value == 0:
            raise ValueError("U cannot be zero.")
        username_password = "%s%s:%s" % (
            self.__pool_id.split("_")[1],
            username,
            password,
        )
        username_password_hash = hash_sha256(username_password.encode("utf-8"))

        x_value = hex_to_long(hex_hash(pad_hex(salt) + username_password_hash))
        g_mod_pow_xn = pow(self.g, x_value, self.big_n)
        int_value2 = server_b_value - self.k * g_mod_pow_xn
        s_value = pow(
            int_value2, self.small_a_value + u_value * x_value, self.big_n
        )
        hkdf = compute_hkdf(
            bytearray.fromhex(pad_hex(s_value)),
            bytearray.fromhex(pad_hex(long_to_hex(u_value))),
        )
        return hkdf

    async def get_auth_params(self):
        auth_params = {
            "USERNAME": self.username,
            "SRP_A": await self.loop.run_in_executor(
                None, long_to_hex, self.large_a_value
            ),
        }
        if self.client_secret is not None:
            auth_params.update(
                {
                    "SECRET_HASH": await self.get_secret_hash(
                        self.username, self.__client_id, self.client_secret
                    )
                }
            )
        return auth_params

    @staticmethod
    async def get_secret_hash(username, client_id, client_secret):
        message = bytearray(username + client_id, "utf-8")
        hmac_obj = hmac.new(
            bytearray(client_secret, "utf-8"), message, hashlib.sha256
        )
        return base64.standard_b64encode(hmac_obj.digest()).decode("utf-8")

    async def process_challenge(self, challenge_parameters):
        self.user_id = challenge_parameters["USER_ID_FOR_SRP"]
        salt_hex = challenge_parameters["SALT"]
        srp_b_hex = challenge_parameters["SRP_B"]
        secret_block_b64 = challenge_parameters["SECRET_BLOCK"]
        # re strips leading zero from a day number (required by AWS Cognito)
        timestamp = re.sub(
            r" 0(\d) ",
            r" \1 ",
            datetime.datetime.utcnow().strftime("%a %b %d %H:%M:%S UTC %Y"),
        )
        hkdf = await self.loop.run_in_executor(
            None,
            self.get_password_authentication_key,
            self.user_id,
            self.password,
            srp_b_hex,
            salt_hex,
        )
        secret_block_bytes = base64.standard_b64decode(secret_block_b64)
        msg = (
            bytearray(self.__pool_id.split("_")[1], "utf-8")
            + bytearray(self.user_id, "utf-8")
            + bytearray(secret_block_bytes)
            + bytearray(timestamp, "utf-8")
        )
        hmac_obj = hmac.new(hkdf, msg, digestmod=hashlib.sha256)
        signature_string = base64.standard_b64encode(hmac_obj.digest())
        response = {
            "TIMESTAMP": timestamp,
            "USERNAME": self.user_id,
            "PASSWORD_CLAIM_SECRET_BLOCK": secret_block_b64,
            "PASSWORD_CLAIM_SIGNATURE": signature_string.decode("utf-8"),
        }
        if self.client_secret is not None:
            response.update(
                {
                    "SECRET_HASH": await self.get_secret_hash(
                        self.username, self.__client_id, self.client_secret
                    )
                }
            )

        return response

    async def login(self):
        """Login into a Hive account."""
        if self.useFile:
            return self.file_response

        await self.async_init()
        boto_client = await self.client
        auth_params = await self.get_auth_params()
        response = None
        result = None
        try:
            response = await self.loop.run_in_executor(
                None,
                functools.partial(
                    boto_client.initiate_auth,
                    AuthFlow="USER_SRP_AUTH",
                    AuthParameters=auth_params,
                    ClientId=self.__client_id,
                ),
            )
        except botocore.exceptions.ClientError as err:
            if err.__class__.__name__ == "UserNotFoundException":
                return "INVALID_USER"
        except botocore.exceptions.EndpointConnectionError as err:
            if err.__class__.__name__ == "EndpointConnectionError":
                return "CONNECTION_ERROR"

        if response["ChallengeName"] == self.PASSWORD_VERIFIER_CHALLENGE:
            challenge_response = await self.process_challenge(
                response["ChallengeParameters"]
            )
            try:
                result = await self.loop.run_in_executor(
                    None,
                    functools.partial(
                        boto_client.respond_to_auth_challenge,
                        ClientId=self.__client_id,
                        ChallengeName=self.PASSWORD_VERIFIER_CHALLENGE,
                        ChallengeResponses=challenge_response,
                    ),
                )
            except botocore.exceptions.ClientError as err:
                if err.__class__.__name__ == "NotAuthorizedException":
                    return "INVALID_PASSWORD"
            except botocore.exceptions.EndpointConnectionError as err:
                if err.__class__.__name__ == "EndpointConnectionError":
                    return "CONNECTION_ERROR"

            return result
        else:
            raise NotImplementedError(
                "The %s challenge is not supported" % response["ChallengeName"]
            )

    async def sms_2fa(self, entered_code, challenge_parameters):
        boto_client = await self.client
        session = challenge_parameters.get("Session")
        code = str(entered_code)
        result = None
        try:
            result = await self.loop.run_in_executor(
                None,
                functools.partial(
                    boto_client.respond_to_auth_challenge,
                    ClientId=self.__client_id,
                    ChallengeName=self.SMS_MFA_CHALLENGE,
                    Session=session,
                    ChallengeResponses={
                        "SMS_MFA_CODE": code,
                        "USERNAME": self.user_id,
                    },
                ),
            )
        except botocore.exceptions.ClientError as err:
            if (
                err.__class__.__name__ == "NotAuthorizedException"
                or err.__class__.__name__ == "CodeMismatchException"
            ):
                return "INVALID_CODE"
        except botocore.exceptions.EndpointConnectionError as err:
            if err.__class__.__name__ == "EndpointConnectionError":
                return "CONNECTION_ERROR"

        return result


def hex_to_long(hex_string):
    return int(hex_string, 16)


def get_random(nbytes):
    random_hex = binascii.hexlify(os.urandom(nbytes))
    return hex_to_long(random_hex)


def hash_sha256(buf):
    """AuthenticationHelper.hash"""
    a = hashlib.sha256(buf).hexdigest()
    return (64 - len(a)) * "0" + a


def hex_hash(hex_string):
    return hash_sha256(bytearray.fromhex(hex_string))


def calculate_u(big_a, big_b):
    """
    Calculate the client's value U which is the hash of A and B
    :param {Long integer} big_a Large A value.
    :param {Long integer} big_b Server B value.
    :return {Long integer} Computed U value.
    """
    u_hex_hash = hex_hash(pad_hex(big_a) + pad_hex(big_b))
    return hex_to_long(u_hex_hash)


def long_to_hex(long_num):
    return "%x" % long_num


def pad_hex(long_int):
    """
    Converts a Long integer (or hex string) \
        to hex format padded with zeroes for hashing
    :param {Long integer|String} long_int Number or string to pad.
    :return {String} Padded hex string.
    """
    if not isinstance(long_int, six.string_types):
        hash_str = long_to_hex(long_int)
    else:
        hash_str = long_int
    if len(hash_str) % 2 == 1:
        hash_str = "0%s" % hash_str
    elif hash_str[0] in "89ABCDEFabcdef":
        hash_str = "00%s" % hash_str
    return hash_str


def compute_hkdf(ikm, salt):
    """
    Standard hkdf algorithm
    :param {Buffer} ikm Input key material.
    :param {Buffer} salt Salt value.
    :return {Buffer} Strong key material.
    @private
    """
    prk = hmac.new(salt, ikm, hashlib.sha256).digest()
    info_bits_update = info_bits + bytearray(chr(1), "utf-8")
    hmac_hash = hmac.new(prk, info_bits_update, hashlib.sha256).digest()
    return hmac_hash[:16]



"""Hive API Module."""
import json
import operator
from datetime import datetime
from typing import Optional

import urllib3
from aiohttp import ClientResponse, ClientSession, web_exceptions
from pyquery import PyQuery

from pyhiveapi.helper.const import HTTP_UNAUTHORIZED

from ..helper.hive_data import Data
from ..helper.hive_exceptions import FileInUse, NoApiToken
from ..helper.logger import Logger

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class HiveAsync:
    """Hive API Code."""

    def __init__(self, websession: Optional[ClientSession] = None):
        """Hive API initialisation."""
        self.logger = Logger()
        self.baseUrl = "https://beekeeper.hivehome.com/1.0"
        self.urls = {
            "properties": "https://sso.hivehome.com/",
            "login": self.baseUrl + "/cognito/login",
            "refresh": self.baseUrl + "/cognito/refresh-token",
            "long_lived": "https://api.prod.bgchprod.info/omnia/accessTokens",
            "weather": "https://weather.prod.bgchprod.info/weather",
            "holiday_mode": "/holiday-mode",
            "all": self.baseUrl
            + "/nodes/all?products=true&devices=true&actions=true",
            "devices": self.baseUrl + "/devices",
            "products": self.baseUrl + "/products",
            "actions": self.baseUrl + "/actions",
            "nodes": self.baseUrl + "/nodes/{0}/{1}",
        }
        self.headers = {
            "content-type": "application/json",
            "Accept": "*/*",
        }
        self.timeout = 10
        self.json_return = {
            "original": "No response to Hive API request",
            "parsed": "No response to Hive API request",
        }
        websession = Data.haWebsession if websession is None else websession
        self.websession = ClientSession() if websession is None else websession

    async def request(self, method: str, url: str, **kwargs) -> ClientResponse:
        """Make a request."""
        data = kwargs.get("data", None)
        await self.logger.log(
            "No_ID",
            "API",
            "Request is - {0}:{1}  Body is {2}",
            info=[method, url, data],
        )

        try:
            self.headers.update({"authorization": Data.tokens["token"]})
        except KeyError:
            if "sso" in url:
                pass
            else:
                self.logger.log("No_ID", "API", "ERROR - NO API TOKEN")
                raise NoApiToken

        async with self.websession.request(
            method, url, headers=self.headers, data=data
        ) as resp:
            await resp.json(content_type=None)
            self.json_return.update({"original": resp.status})
            self.json_return.update(
                {"parsed": await resp.json(content_type=None)}
            )

        if operator.contains(str(resp.status), "20"):
            await self.logger.log(
                "API", "API", "Response is - {0}", info=[str(resp.status)]
            )
        elif resp.status == HTTP_UNAUTHORIZED:
            await self.logger.LOGGER.error(
                f"Hive token as expired when calling {url} - "
                f"HTTP status is - {resp.status}"
            )
        else:
            await self.logger.LOGGER.error(
                f"Something has gone wrong calling {url} - "
                f"HTTP status is - {resp.status}"
            )

    async def refreshTokens(self):
        """Refresh tokens"""
        url = self.urls["refresh"]
        jsc = (
            "{"
            + ",".join(
                (
                    '"' + str(i) + '": ' '"' + str(t) + '" '
                    for i, t in Data.tokens.items()
                )
            )
            + "}"
        )
        try:
            await self.request("post", url, data=jsc)

            if self.json_return["original"] == 200:
                info = self.json_return["parsed"]
                if "token" in info:
                    Data.tokens.update({"token": info["token"]})
                    Data.tokens.update({"refreshToken": info["refreshToken"]})
                    Data.tokens.update({"accessToken": info["accessToken"]})

                    self.urls.update({"base": info["platform"]["endpoint"]})
                    self.urls.update(
                        {"camera": info["platform"]["cameraPlatform"]}
                    )
                    Data.tokenCreated = datetime.now()
                return True
        except (ConnectionError, IOError, RuntimeError, ZeroDivisionError):
            await self.error()

        return self.json_return

    async def getLoginInfo(self):
        """Get login properties to make the login request."""
        url = self.urls["properties"]
        try:
            data = await self.request("get", url)
            html = PyQuery(data.content)
            json_data = json.loads(
                '{"'
                + (html("script:first").text())
                .replace(",", ', "')
                .replace("=", '":')
                .replace("window.", "")
                + "}"
            )

            Data.loginData.update({"UPID": json_data["HiveSSOPoolId"]})
            Data.loginData.update(
                {"CLIID": json_data["HiveSSOPublicCognitoClientId"]}
            )
            Data.loginData.update({"REGION": json_data["HiveSSOPoolId"]})
            return Data.loginData
        except (IOError, RuntimeError, ZeroDivisionError):
            await self.error()

    async def getAll(self):
        """Build and query all endpoint."""
        url = self.urls["all"]
        try:
            await self.request("get", url)
        except (IOError, RuntimeError, ZeroDivisionError):
            await self.error()

        return self.json_return

    async def getDevices(self):
        """Call the get devices endpoint."""
        url = self.urls["devices"]
        try:
            await self.request("get", url)
        except (IOError, RuntimeError, ZeroDivisionError):
            await self.error()

        return self.json_return

    async def get_products(self, token):
        """Call the get products endpoint."""
        url = self.urls["products"]
        try:
            await self.request("get", url)
        except (IOError, RuntimeError, ZeroDivisionError):
            await self.error()

        return self.json_return

    async def get_actions(self):
        """Call the get actions endpoint."""
        url = self.urls["actions"]
        try:
            await self.request("get", url)
        except (IOError, RuntimeError, ZeroDivisionError):
            await self.error()

        return self.json_return

    async def motion_sensor(self, token, sensor, fromepoch, toepoch):
        """Call a way to get motion sensor info."""
        url = (
            self.urls["base"]
            + self.urls["products"]
            + "/"
            + sensor["type"]
            + "/"
            + sensor["id"]
            + "/events?from="
            + str(fromepoch)
            + "&to="
            + str(toepoch)
        )
        try:
            await self.request("get", url)
        except (IOError, RuntimeError, ZeroDivisionError):
            await self.error()

        return self.json_return

    async def get_weather(self, weather_url):
        """Call endpoint to get local weather from Hive API."""
        t_url = self.urls["weather"] + weather_url
        url = t_url.replace(" ", "%20")
        try:
            await self.request("get", url)
        except (IOError, RuntimeError, ZeroDivisionError, ConnectionError):
            await self.error()

        return self.json_return

    async def set_state(self, n_type, n_id, **kwargs):
        """Set the state of a Device."""
        jsc = (
            "{"
            + ",".join(
                (
                    '"' + str(i) + '": ' '"' + str(t) + '" '
                    for i, t in kwargs.items()
                )
            )
            + "}"
        )

        url = self.urls["nodes"].format(n_type, n_id)
        try:
            await self.is_file_being_used()
            await self.request("post", url, data=jsc)
        except (FileInUse, IOError, RuntimeError, ConnectionError) as e:
            if e.__class__.__name__ == "FileInUse":
                return {"original": "file"}
            else:
                await self.error()

        return self.json_return

    async def set_action(self, n_id, data):
        """Set the state of a Action."""
        jsc = data
        url = self.urls["base"] + self.urls["actions"] + "/" + n_id
        try:
            await self.is_file_being_used()
            await self.request("put", url, data=jsc)
        except (FileInUse, IOError, RuntimeError, ConnectionError) as e:
            if e.__class__.__name__ == "FileInUse":
                return {"original": "file"}
            else:
                await self.error()

        return self.json_return

    async def error(self):
        """An error has occured iteracting wth the Hive API."""
        await self.logger.log(
            "API_ERROR", "ERROR", "Error attempting API call"
        )
        raise web_exceptions.HTTPError

    async def is_file_being_used(self):
        """Check if running in file mode."""
        if Data.file:
            raise FileInUse()



import os
import re

from setuptools import find_packages, setup


def requirements_from_file(filename="requirements.txt"):
    with open(os.path.join(os.path.dirname(__file__), filename)) as r:
        reqs = r.read().strip().split("\n")
    # Return non emtpy lines and non comments
    return [r for r in reqs if re.match(r"^\w+", r)]


setup(
    name="pyhiveapi",
    version="0.3.4.5",
    description="A Python library to interface with the Hive API",
    long_description="A Python library to interface with the Hive API",
    url="https://github.com/Pyhive/pyhiveapi",
    package_data={"pyhiveapi.pyhiveapi": ["*.json"]},
    include_package_data=True,
    author="Rendili",
    author_email="rendili@outlook.com",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        # "Topic:: Software Development:: Libraries:: Python Modules",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires=">=3.6.*",
    keywords="Hive API Library",
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    entry_points={"console_scripts": ["hive=pyhiveapi.hive:Hive"]},
    install_requires=requirements_from_file(),
    extras_require={"dev": requirements_from_file("requirements_test.txt")},
)



""""Hive Hotwater Module. """
from .helper.hive_data import Data
from .session import Session


class Hotwater(Session):
    """Hive Hotwater Code."""

    hotwaterType = "Hotwater"

    async def get_hotwater(self, device):
        """Get light data."""
        await self.logger.log(
            device["hiveID"], self.hotwaterType, "Getting hot water data."
        )
        device["deviceData"].update(
            {"online": await self.attr.online_offline(device["device_id"])}
        )

        if device["deviceData"]["online"]:

            dev_data = {}
            self.helper.deviceRecovered(device["device_id"])
            data = Data.devices[device["device_id"]]
            dev_data = {
                "hiveID": device["hiveID"],
                "hiveName": device["hiveName"],
                "hiveType": device["hiveType"],
                "haName": device["haName"],
                "haType": device["haType"],
                "device_id": device["device_id"],
                "device_name": device["device_name"],
                "status": {"current_operation": await self.get_mode(device)},
                "deviceData": data.get("props", None),
                "parentDevice": data.get("parent", None),
                "custom": device.get("custom", None),
                "attributes": await self.attr.state_attributes(
                    device["device_id"], device["hiveType"]
                ),
            }

            await self.logger.log(
                device["hiveID"],
                self.hotwaterType,
                "Device update {0}",
                info=dev_data["status"],
            )
            Data.ha_devices.update({device["hiveID"]: dev_data})
            return Data.ha_devices[device["hiveID"]]
        else:
            await self.logger.error_check(
                device["device_id"], "ERROR", device["deviceData"]["online"]
            )
            return device

    async def get_mode(self, device):
        """Get hotwater current mode."""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["state"]["mode"]
            if state == "BOOST":
                state = data["props"]["previous"]["mode"]
            final = Data.HIVETOHA[self.hotwaterType].get(state, state)
        except KeyError as e:
            await self.logger.error(e)

        return final

    @staticmethod
    async def get_operation_modes():
        """Get heating list of possible modes."""
        return ["SCHEDULE", "ON", "OFF"]

    async def get_boost(self, device):
        """Get hot water current boost status."""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["state"]["boost"]
            final = Data.HIVETOHA["Boost"].get(state, "ON")
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def get_boost_time(self, device):
        """Get hotwater boost time remaining."""
        state = None
        if await self.get_boost(device) == "ON":
            try:
                data = Data.products[device["hiveID"]]
                state = data["state"]["boost"]
            except KeyError as e:
                await self.logger.error(e)

        return state

    async def get_state(self, device):
        """Get hot water current state."""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["state"]["status"]
            mode_current = await self.get_mode(device)
            if mode_current == "SCHEDULE":
                if await self.get_boost(device) == "ON":
                    state = "ON"
                else:
                    snan = await self.helper.getScheduleNNL(
                        data["state"]["schedule"]
                    )
                    state = snan["now"]["value"]["status"]

            final = Data.HIVETOHA[self.hotwaterType].get(state, state)
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def get_schedule_now_next_later(self, device):
        """Hive get hotwater schedule now, next and later."""
        state = None

        try:
            mode_current = await self.get_mode(device)
            if mode_current == "SCHEDULE":
                data = Data.products[device["hiveID"]]
                state = await self.helper.getScheduleNNL(
                    data["state"]["schedule"]
                )
        except KeyError as e:
            await self.logger.error(e)

        return state

    async def set_mode(self, device, new_mode):
        """Set hot water mode."""
        final = False

        if device["hiveID"] in Data.products:
            await self.hiveRefreshTokens()
            data = Data.products[device["hiveID"]]
            resp = await self.api.set_state(
                data["type"], device["hiveID"], mode=new_mode
            )
            if resp["original"] == 200:
                final = True
                await self.getDevices(device["hiveID"])

        return final

    async def turn_boost_on(self, device, mins):
        """Turn hot water boost on."""
        final = False

        if (
            mins > 0
            and device["hiveID"] in Data.products
            and device["deviceData"]["online"]
        ):
            await self.hiveRefreshTokens()
            data = Data.products[device["hiveID"]]
            resp = await self.api.set_state(
                data["type"], device["hiveID"], mode="BOOST", boost=mins
            )
            if resp["original"] == 200:
                final = True
                await self.getDevices(device["hiveID"])

        return final

    async def turn_boost_off(self, device):
        """Turn hot water boost off."""
        final = False

        if (
            device["hiveID"] in Data.products
            and await self.get_boost(device) == "ON"
            and device["deviceData"]["online"]
        ):
            await self.hiveRefreshTokens()
            data = Data.products[device["hiveID"]]
            prev_mode = data["props"]["previous"]["mode"]
            resp = await self.api.set_state(
                data["type"], device["hiveID"], mode=prev_mode
            )
            if resp["original"] == 200:
                await self.getDevices(device["hiveID"])
                final = True

        return final



"""Hive API Module."""
import json

import requests
import urllib3
from pyquery import PyQuery

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class HiveApi:
    """Hive API Code."""

    def __init__(self):
        """Hive API initialisation."""
        self.urls = {
            "properties": "https://sso.hivehome.com/",
            "login": "https://beekeeper.hivehome.com/1.0/cognito/login",
            "refresh": "https://beekeeper.hivehome.com/1.0/cognito/refresh-token",
            "long_lived": "https://api.prod.bgchprod.info/omnia/accessTokens",
            "base": "https://beekeeper-uk.hivehome.com/1.0",
            "weather": "https://weather.prod.bgchprod.info/weather",
            "holiday_mode": "/holiday-mode",
            "all": "/nodes/all?products=true&devices=true&actions=true",
            "devices": "/devices",
            "products": "/products",
            "actions": "/actions",
            "nodes": "/nodes/{0}/{1}",
        }
        self.headers = {
            "content-type": "application/json",
            "Accept": "*/*",
            "authorization": "None",
        }
        self.timeout = 10
        self.json_return = {
            "original": "No response to Hive API request",
            "parsed": "No response to Hive API request",
        }

    def refreshTokens(self, tokens):
        """Get new session tokens"""
        url = self.urls["refresh"]
        jsc = (
            "{"
            + ",".join(
                (
                    '"' + str(i) + '": ' '"' + str(t) + '" '
                    for i, t in tokens.items()
                )
            )
            + "}"
        )
        try:
            response = requests.post(
                url=url, headers=self.headers, data=jsc, timeout=self.timeout
            )
            self.json_return.update({"original": response.status_code})
            self.json_return.update({"parsed": response.json()})
        except (IOError, RuntimeError, ZeroDivisionError):
            self.error()

        return self.json_return

    def getLoginInfo(self):
        """Get login properties to make the login request."""
        url = self.urls["properties"]
        try:
            data = requests.get(url=url, verify=False, timeout=self.timeout)
            html = PyQuery(data.content)
            json_data = json.loads(
                '{"'
                + (html("script:first").text())
                .replace(",", ', "')
                .replace("=", '":')
                .replace("window.", "")
                + "}"
            )

            loginData = {}
            loginData.update({"UPID": json_data["HiveSSOPoolId"]})
            loginData.update(
                {"CLIID": json_data["HiveSSOPublicCognitoClientId"]}
            )
            loginData.update({"REGION": json_data["HiveSSOPoolId"]})
            return loginData
        except (IOError, RuntimeError, ZeroDivisionError):
            self.error()

    def getAllData(self, session_id):
        """Build and query all endpoint."""
        self.headers.update({"authorization": session_id})
        url = self.urls["base"] + self.urls["all"]
        try:
            jsc = None
            response = requests.get(
                url=url, headers=self.headers, data=jsc, timeout=self.timeout
            )
            self.json_return.update({"original": response.status_code})
            self.json_return.update({"parsed": response.json()})
        except (IOError, RuntimeError, ZeroDivisionError):
            self.error()

        return self.json_return

    def getDevices(self, session_id):
        """Call the get devices endpoint."""
        self.headers.update({"authorization": session_id})
        url = self.urls["base"] + self.urls["devices"]
        try:
            jsc = None
            response = requests.get(
                url=url, headers=self.headers, data=jsc, timeout=self.timeout
            )
            self.json_return.update({"original": response.status_code})
            self.json_return.update({"parsed": response.json()})
        except (IOError, RuntimeError, ZeroDivisionError):
            self.error()

        return self.json_return

    def get_products(self, session_id):
        """Call the get products endpoint."""
        self.headers.update({"authorization": session_id})
        url = self.urls["base"] + self.urls["products"]
        try:
            jsc = None
            response = requests.get(
                url=url, headers=self.headers, data=jsc, timeout=self.timeout
            )
            self.json_return.update({"original": response.status_code})
            self.json_return.update({"parsed": response.json()})
        except (IOError, RuntimeError, ZeroDivisionError):
            self.error()

        return self.json_return

    def get_actions(self, session_id):
        """Call the get actions endpoint."""
        self.headers.update({"authorization": session_id})
        url = self.urls["base"] + self.urls["actions"]
        try:
            jsc = None
            response = requests.get(
                url=url, headers=self.headers, data=jsc, timeout=self.timeout
            )
            self.json_return.update({"original": response.status_code})
            self.json_return.update({"parsed": response.json()})
        except (IOError, RuntimeError, ZeroDivisionError):
            self.error()

        return self.json_return

    def motion_sensor(self, session_id, sensor, fromepoch, toepoch):
        """Call a way to get motion sensor info."""
        self.headers.update({"authorization": session_id})
        url = (
            self.urls["base"]
            + self.urls["products"]
            + "/"
            + sensor["type"]
            + "/"
            + sensor["id"]
            + "/events?from="
            + str(fromepoch)
            + "&to="
            + str(toepoch)
        )
        try:
            jsc = None
            response = requests.get(
                url=url, headers=self.headers, data=jsc, timeout=self.timeout
            )
            self.json_return.update({"original": response.status_code})
            self.json_return.update({"parsed": response.json()})
        except (IOError, RuntimeError, ZeroDivisionError):
            self.error()

        return self.json_return

    def get_weather(self, weather_url):
        """Call endpoint to get local weather from Hive API."""
        t_url = self.urls["weather"] + weather_url
        url = t_url.replace(" ", "%20")
        try:
            jsc = None
            response = requests.get(
                url=url, headers=self.headers, data=jsc, timeout=self.timeout
            )
            self.json_return.update({"original": response.status_code})
            self.json_return.update({"parsed": response.json()})
        except (IOError, RuntimeError, ZeroDivisionError, ConnectionError):
            self.error()

        return self.json_return

    def set_state(self, session_id, n_type, n_id, **kwargs):
        """Set the state of a Device."""
        self.headers.update({"authorization": session_id})
        jsc = (
            "{"
            + ",".join(
                (
                    '"' + str(i) + '": ' '"' + str(t) + '" '
                    for i, t in kwargs.items()
                )
            )
            + "}"
        )

        url = self.urls["base"] + self.urls["nodes"].format(n_type, n_id)

        try:
            response = requests.post(
                url=url, headers=self.headers, data=jsc, timeout=self.timeout
            )
            self.json_return.update({"original": response.status_code})
            self.json_return.update({"parsed": response.json()})
        except (IOError, RuntimeError, ZeroDivisionError, ConnectionError):
            self.error()

        return self.json_return

    def set_action(self, session_id, n_id, data):
        """Set the state of a Action."""
        self.headers.update({"authorization": session_id})
        jsc = data
        url = self.urls["base"] + self.urls["actions"] + "/" + n_id
        try:
            response = requests.put(
                url=url, headers=self.headers, data=jsc, timeout=self.timeout
            )
            self.json_return.update({"original": response.status_code})
            self.json_return.update({"parsed": response.json()})
        except (IOError, RuntimeError, ZeroDivisionError, ConnectionError):
            self.error()

        return self.json_return

    def error(self):
        """An error has occured iteracting wth the Hive API."""
        self.json_return.update({"original": "Error making API call"})
        self.json_return.update({"parsed": "Error making API call"})



from typing import Optional

from aiohttp import ClientSession

from .action import Action
from .heating import Heating
from .helper.hive_data import Data
from .hotwater import Hotwater
from .hub import Hub
from .light import Light
from .plug import Plug
from .sensor import Sensor
from .session import Session


class Hive:
    """Hive Class."""

    def __init__(self, websession: Optional[ClientSession] = None) -> None:
        Data.haWebsession = websession
        self.session = Session()
        self.action = Action()
        self.heating = Heating()
        self.hotwater = Hotwater()
        self.hub = Hub()
        self.light = Light()
        self.switch = Plug()
        self.sensor = Sensor()
        self.devices = {}



"""Hive Hub Module."""
from .helper.hive_data import Data
from .session import Session


class Hub(Session):
    """ Hive Hub Code. """

    hubType = "Hub"
    logType = "Sensor"

    async def hub_smoke(self, device):
        """Get the online status of the Hive hub."""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["props"]["sensors"]["SMOKE_CO"]["active"]
            final = Data.HIVETOHA[self.hubType]["Smoke"].get(state, state)
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def hub_dog_bark(self, device):
        """Get the online status of the Hive hub."""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["props"]["sensors"]["DOG_BARK"]["active"]
            final = Data.HIVETOHA[self.hubType]["Dog"].get(state, state)
        except KeyError as e:
            await self.logger.error(e)

        return final

    async def hub_glass(self, device):
        """Get the glass detected status from the Hive hub."""
        state = None
        final = None

        try:
            data = Data.products[device["hiveID"]]
            state = data["props"]["sensors"]["GLASS_BREAK"]["active"]
            final = Data.HIVETOHA[self.hubType]["Glass"].get(state, state)
        except KeyError as e:
            await self.logger.error(e)

        return final
